# Import modules
import json
import os
import shutil
import sqlite3
from regulome_app.webapp.app import app  # , db
from regulome_app.config import logger, configs
from regulome_app.webapp import models as models
from flask import render_template, url_for, redirect, flash, session, request, send_from_directory, jsonify
from collections import defaultdict
from regulome_app.webapp.tools.parse_messages import ParseMessage
from regulome_app.webapp.tools.plot import Plot
from regulome_app.webapp.tools.uploads import Upload
from werkzeug.utils import secure_filename


# Global variables
__author__ = 'Loris Mularoni'

genes_coords = {}
for build in models.BUILDS.keys():
    file_path = os.path.join(configs['data']['data'], build, '{}_UCSC_RefSeq_genes_simple.txt'.format(build))
    genes_coords[build] = defaultdict(list)
    with open(file_path, 'r') as fi:
        for line in fi:
            line = line.strip().split('\t')
            gene, chromosome, strand, start, end, _ = line
            gene = gene.upper()
            genes_coords[build][gene].append((gene, chromosome, start, end))


@app.context_processor
def inject_globals():
    deploy_mode = configs['deploy_mode']['configuration']
    return dict(
        deploy_mode=deploy_mode if deploy_mode in ('development', 'testing', 'production') else 'development',
        snps=models.SNPS,
        builds=models.BUILDS,
        chromosomes=models.CHROMOSOMES_SORTED,
        regions=models.REGIONS,
        tfbs=models.TFBS,
        chromatin=models.CHROMATIN,
        ranges=models.RANGES
    )


def reset_session_messages():
    """Reset the session messages to the default values"""
    session['messages'] = json.dumps(models.default_messages)


def update_messages(new_messages):
    """Update the information of messages"""
    messages = json.loads(session['messages'])
    for key, value in new_messages.items():
        messages[key] = value
    session['messages'] = json.dumps(messages)


@app.route('/')
def root():
    return redirect(url_for('intro'))


@app.route('/home', methods=['GET', 'POST'])
def intro():
    """Show the intro page or redirect to another page"""
    if 'username' not in session:
        session['username'] = os.urandom(24).hex()
    if 'messages' not in session:
        reset_session_messages()
    if request.form.get('action', None) == 'submit':
        messages = ParseMessage(messages=json.dumps(request.form))
        session['messages'] = messages.get_json()
        return redirect(url_for('visualize_plot'))
    elif request.form.get('action', None) == 'upload':
        update_messages(request.form)
        # The argument code is to force a POST redirect
        return redirect(url_for('upload_file'), code=307)
    else:
        # Restore the previous options
        return render_template('intro.html', messages=json.loads(session['messages']))


@app.route('/submit', methods=['GET', 'POST'])
def submit():
    """Parse the submission of the form"""
    if request.form.get('navbar_link', None) in ('home', 'home_main'):
        return redirect(url_for('intro'))
    elif request.form.get('navbar_link', None) == 'data':
        return redirect(url_for('data'))
    elif request.form.get('navbar_link', None) == 'credits':
        return redirect(url_for('credits'))
    elif request.form.get('navbar_link', None) == 'contact':
        return redirect(url_for('contact'))
    elif request.form.get('navbar_link', None) == 'info':
        return redirect(url_for('info'))
    elif request.form.get('navbar_link', None) == 'upload':
        update_messages(request.form)
        # The argument code is to force a POST redirect
        return redirect(url_for('upload_file'), code=307)
    elif request.form.get('navbar_link', None) in ('clear_region_upload', 'clear_snp_upload'):
        messages = json.loads(session['messages'])
        if request.form.get('navbar_link') == 'clear_region_upload':
            messages['upload_regions'] = 'None'
            messages['shared_regions'] = 'None'
        else:
            messages['upload_snps'] = 'None'
            messages['shared_snps'] = 'None'
        session['messages'] = json.dumps(messages)
        return redirect(url_for('intro'))
    else:
        try:
            messages = ParseMessage(messages=json.dumps(request.form))
            session['messages'] = messages.get_json()
            logger.debug(session['messages'])
            return redirect(url_for('visualize_plot'))
        except KeyError as e:
            return redirect(url_for('intro'))


@app.route('/show')
def visualize_plot(messages=None):
    """Create and visualize the plot"""
    if messages is None:
        # Redirect to the home page
        if 'messages' not in session:
            return redirect(url_for('intro'))

        messages = json.loads(session['messages'])
        ip = request.environ.get('HTTP_X_FORWARDED_FOR', request.remote_addr)
        plot = Plot(messages=messages, username=session['username'], ip=ip)
        plot.run()
        messages['file_path'] = plot.get_results_path()
        messages['file_name'] = plot.get_results_name()
        messages['downloads'] = enable_downloads(messages['file_path'])
        messages['table_tfbs'] = table_tfbs(messages['file_path'], messages['region'])
        messages['table_snp'] = table_snp(messages['file_path'])
        messages['table_expression'] = table_expression(messages['file_path'])

    return render_template('plot.html', messages=messages)


@app.route('/cache/<path:filename>')
def cache(filename):
    """Return the path of a file in the cache folder"""
    return send_from_directory(configs['output']['cache'], filename, as_attachment=True)


@app.route('/api', methods=['GET'])
def retrieve_genes():
    """Retrieve the genes matching the gene_name"""
    build = request.args.get('build', None)
    gene_name = request.args.get('gene_name', None)
    fuzzy = eval(request.args.get('fuzzy', None))

    # sql_file = os.path.join(configs['data']['data'], build, '{}_genes.db'.format(build))
    # genes = query_sql(sql_file, gene_name)
    # genes = [gene for gene in genes if gene[6] == 'gene']
    # gene_file = os.path.join(configs['data']['data'], build, '{}_UCSC_RefSeq_genes_simple.txt'.format(build))
    # genes = query_file(gene_file, gene_name, fuzzy)
    if fuzzy:
        genes = []
        for gene in genes_coords[build].keys():
            if gene.startswith(gene_name.upper()):
                for isoform in genes_coords[build][gene]:
                    genes.append(isoform)
        genes = sorted(genes)[:6]
    else:
        genes = genes_coords[build][gene_name] if gene_name in genes_coords[build].keys() else []
    return jsonify({'genes': genes, 'fuzzy': fuzzy})


def query_file(file_path, symbol, fuzzy):
    """Return up to 10 lines of the file where the
    initials of the gene correspond to symbol.
    """
    genes = []
    with open(file_path, 'r') as fi:
        for line in fi:
            line = line.strip().split('\t')
            gene, chromosome, strand, start, end, _ = line
            gene = gene.upper()
            if fuzzy and gene.startswith(symbol.upper()):
                genes.append((gene, chromosome, start, end))
            elif not fuzzy and gene == symbol.upper():
                return gene.upper(), chromosome, start, end
    return genes[:10]


def query_sql(db, symbol):
    """Queries a sqlite database. Return up to
    10 genes whose initials correspond to symbol
    """
    conn = sqlite3.connect(db)
    c = conn.cursor()
    query = "SELECT * FROM {} WHERE SYMBOL LIKE '{}%' LIMIT 10;".format('genes', symbol)
    c.execute(query)
    result = c.fetchall()
    conn.close()
    return result


@app.route('/share', methods=['GET'])
def share():
    """Load the shared datasets"""
    if 'username' not in session:
        session['username'] = os.urandom(24).hex()
    if 'messages' not in session:
        reset_session_messages()

    messages = json.loads(session['messages'])

    # Copy the files in the user space
    for id_, name_, shared_, upload_ in zip(
            ('regions_id', 'snps_id'),
            ('regions_name', 'snps_name'),
            ('shared_regions', 'shared_snps'),
            ('upload_regions', 'upload_snps')):
        source = os.path.join(
            app.config['UPLOAD_FOLDER'],
            request.args.get(id_, 'None')
        )
        destination = os.path.join(
            app.config['UPLOAD_FOLDER'],
            "{}_{}".format(
                session['username'],
                request.args.get(name_, 'None')
            )
        )
        try:
            shutil.copyfile(source, destination)
        except FileNotFoundError as e:
            continue
        messages[shared_] = request.args.get(id_, 'None')
        messages[upload_] = request.args.get(name_, 'None')

    session['messages'] = json.dumps(messages)

    return redirect(url_for('intro'))


@app.route('/credits')
def credits():
    """Show the credit page"""
    return render_template('credits.html')


@app.route('/contact')
def contact():
    """Show the contact page"""
    return render_template('contact.html')


@app.route('/info')
def info():
    """Show the info page"""
    return render_template('info.html')


@app.route('/data')
def data():
    """Show the data page"""
    return render_template('data.html')


@app.route('/show/<key>')
def get_message(key):
    return render_template('intro.html',
                           message=models.MESSAGES.get(key) or "{} not found!".format(key))


@app.route('/add/<key>/<message>')
def add_or_update_message(key, message):
    models.MESSAGES[key] = message
    return render_template('intro.html', message="{} Added/Updated".format(key))


def enable_downloads(file_path):
    """Calculates which files could be downloaded"""
    downloads = {}
    for name, extensions in zip(('sites', 'snps', 'expression', 'figures'),
                                (['.txt'], ['.snp.txt'], ['.expression.txt'], ['.png', '.pdf'])):
        if all([os.path.exists(os.path.join(configs['output']['cache'], file_path + extension))
                for extension in extensions]):
            downloads[name] = 'enabled'
        else:
            downloads[name] = 'disabled'
    return downloads


@app.route('/downloads', methods=['POST'])
def downloads():
    """Return the path of a file in the cache folder"""
    if request.form.get('download', None) == 'download_table':
        filepath = request.form.get('file_path') + '.txt'
        filename = request.form.get('file_name') + '.txt'
        return send_from_directory(
            configs['output']['cache'],
            filepath,
            attachment_filename=filename,
            as_attachment=True
        )
    elif request.form.get('download', None) == 'download_snps':
        filepath = request.form.get('file_path') + '.snp.txt'
        filename = request.form.get('file_name') + '.snp.txt'
        return send_from_directory(
            configs['output']['cache'],
            filepath,
            attachment_filename=filename,
            as_attachment=True
        )
    elif request.form.get('download', None) == 'download_expression':
        filepath = request.form.get('file_path') + '.expression.txt'
        filename = request.form.get('file_name') + '.expression.txt'
        return send_from_directory(
            configs['output']['cache'],
            filepath,
            attachment_filename=filename,
            as_attachment=True
        )
    elif request.form.get('download', None) == 'download_pdf':
        filepath = request.form.get('file_path') + '.pdf'
        filename = request.form.get('file_name') + '.pdf'

        return send_from_directory(
            configs['output']['cache'],
            filepath,
            attachment_filename=filename,
            as_attachment=True
        )
    elif request.form.get('download', None) == 'download_png':
        filepath = request.form.get('file_path') + '.png'
        filename = request.form.get('file_name') + '.png'
        return send_from_directory(
            configs['output']['cache'],
            filepath,
            attachment_filename=filename,
            as_attachment=True
        )
    else:
        flash('An error occurred.')
        return render_template('plot.html', messages=request.form)


def table_tfbs(file_path, region):
    """Return the elements of an html table"""
    filename = os.path.join(configs['output']['cache'], file_path + '.txt')
    table = []

    try:
        tfs = None
        with open(filename, 'r') as fi:
            for i, line in enumerate(fi):
                # Skip the header
                if i == 0:
                    continue
                if region == 'adult':
                    chromosome, start, end, group, tfs, _ = line.split("\t")
                elif region == 'adultStretch':
                    chromosome, start, end, group, _ = line.split("\t")
                elif region == 'progenitor':
                    chromosome, start, group, end, _ = line.split("\t")

                color = models.COLORS[region][group]
                font_color = "white" if color == (0, 0, 0) else "black"
                table.append(dict())
                table[-1]['color'] = color
                table[-1]['font_color'] = font_color
                table[-1]['group'] = group
                table[-1]['chromosome'] = models.CHROMOSOMES_DICT_R[chromosome]
                table[-1]['start'] = start
                table[-1]['end'] = end

                if tfs is not None:
                    table[-1]['tfs'] = tfs

    except FileNotFoundError as e:
        with open(configs['logs']['regulome_log'], 'a') as log:
            log.write("FileNotFoundError in table: {}".format(filename))
    except Exception as e:
        with open(configs['logs']['regulome_log'], 'a') as log:
            log.write("Exception in table: {}".format(filename))
            log.write("Exception is: {}".format(e))
    finally:
        return table


def table_snp(file_path, max_elements=10):
    """Return the most significant 10 elements of an html table"""
    filename = os.path.join(configs['output']['cache'], file_path + '.snp.txt')
    table = []

    try:
        with open(filename, 'r') as fi:
            for i, line in enumerate(fi):
                # Skip the header
                if i == 0:
                    continue
                elif i > max_elements:
                    break
                snp_id, chromosome, position, p_value, dataset, _ = line.split("\t")
                color = 'white'  # models.COLORS[region][group]
                font_color = "white" if color == (0, 0, 0) else "black"
                table.append(dict())
                table[-1]['color'] = color
                table[-1]['font_color'] = font_color
                table[-1]['snp_id'] = snp_id
                table[-1]['chromosome'] = models.CHROMOSOMES_DICT_R[chromosome]
                table[-1]['position'] = position
                table[-1]['p_value'] = p_value
                table[-1]['dataset'] = dataset

    except FileNotFoundError as e:
        with open(configs['logs']['regulome_log'], 'a') as log:
            log.write("FileNotFoundError in table: {}".format(filename))
    except Exception as e:
        with open(configs['logs']['regulome_log'], 'a') as log:
            log.write("Exception in table: {}".format(filename))
            log.write("Exception is: {}".format(e))
    finally:
        return table


def table_expression(file_path):
    """Return the most significant 10 elements of an html table"""
    filename = os.path.join(configs['output']['cache'], file_path + '.expression.txt')
    table = []

    try:
        with open(filename, 'r') as fi:
            for i, line in enumerate(fi):
                # Skip the header
                if i == 0:
                    continue
                chromosome, _, start, end, _, gene, expression, _ = line.split("\t")
                color = 'white'  # models.COLORS[region][group]
                font_color = "white" if color == (0, 0, 0) else "black"
                table.append(dict())
                table[-1]['color'] = color
                table[-1]['font_color'] = font_color
                table[-1]['chromosome'] = models.CHROMOSOMES_DICT_R[chromosome]
                table[-1]['start'] = start
                table[-1]['end'] = end
                table[-1]['gene'] = gene
                table[-1]['expression'] = expression

    except FileNotFoundError as e:
        with open(configs['logs']['regulome_log'], 'a') as log:
            log.write("FileNotFoundError in table: {}".format(filename))
    except Exception as e:
        with open(configs['logs']['regulome_log'], 'a') as log:
            log.write("Exception in table: {}".format(filename))
            log.write("Exception is: {}".format(e))
    finally:
        return table


@app.route('/uploads', methods=['GET', 'POST'])
def upload_file():
    """Save an uploaded file"""
    if request.method == 'POST':
        # Check if the post request has the file part
        if 'region_file' not in request.files and 'snps_file' not in request.files:  # regions_upload, snp_upload
            return redirect(request.url)
        elif request.files['region_file'] == '' and request.files['snps_file'] == '':
            flash('Please select a file to download')
            return redirect(request.url)
        for upload_tag, upload_id, shared_id in zip(
                ('region_file', 'snps_file'),
                ('upload_regions', 'upload_snps'),
                ('shared_regions', 'shared_snps')):
            f = request.files[upload_tag]

            # If the user does not select any file, browser also
            # submits an empty part without filename
            if f.filename == '':
                continue
            if f:  # and allowed_file(file.filename):
                filename = secure_filename(f.filename)
                filepath = os.path.join(app.config['UPLOAD_FOLDER'],
                                        "{}_{}".format(session['username'], filename))
                f.save(filepath + '.raw')

                # Verify the validity of the uploaded file
                upload = Upload(input_file=filepath + '.raw',
                                type_of_file=upload_tag,
                                output_file=filepath)
                if upload.get_errors():
                    flash(upload.get_errors())
                    return redirect(url_for('intro'))
                else:
                    messages = json.loads(session['messages'])
                    messages[upload_id] = filename
                    messages[shared_id] = upload.shared_file
                    session['messages'] = json.dumps(messages)
    return redirect(url_for('intro'))


# class Regions(db.Model):
#     __bind_key__ = 'regions'
#     id = db.Column(db.Integer, primary_key=True)
#     # username = db.Columns(db.String(80), unique=True)





from browser import window

def wrap(func):
    # Transforms a function f into another function that prints a
    # traceback in case of exception
    def f(*args, **kw):
        try:
            return func(*args, **kw)
        except Exception as exc:
            msg = '{0.info}\n{0.__name__}: {0.args[0]}'.format(exc)
            import sys
            sys.stderr.write(msg)
    return f

clear_interval = window.clearInterval
    
clear_timeout = window.clearTimeout

def set_interval(func,interval):
    return window.setInterval(wrap(func),interval)

def set_timeout(func,interval):
    return int(window.setTimeout(wrap(func),interval))

def request_animation_frame(func):
    return int(window.requestAnimationFrame(func))

def cancel_animation_frame(int_id):
    window.cancelAnimationFrame(int_id)

def set_loop_timeout(x):
    # set a variable used to stop loops that last more than x seconds
    assert isinstance(x, int)
    __BRYTHON__.loop_timeout = x


from . import widget
from browser import html, document

class Dialog(html.DIV, widget.DraggableWidget):

    def __init__(self, title):
      html.DIV.__init__(self,
         Class="ui-dialog ui-widget ui-widget-content ui-corner-all ui-front ui-draggable ui-resizable",
         style={'position': 'absolute', 'height': 'auto', 'width': '300px',
                'top': '98px', 'left': '140px', 'display': 'block', 'zIndex':0})

      widget.DraggableWidget.__init__(self, self, 'dialog', id)

      titlebar = html.DIV(Id="titlebar", Class="ui-dialog-titlebar")

      self._title = html.DIV(title, Class="ui-dialog-title")
        
      titlebar <= self._title

      title_button = html.BUTTON('X', Title="close", Class="ui-close-button")

      def dialog_close(e):
          #del document[self._div_shell.id]
          document.remove(self)

      title_button.bind('click', dialog_close)

      titlebar <= title_button

      self.body=html.DIV(Class="ui-dialog-body",
           style={'width': 'auto', 'min-height': '105px', 
                  'max-height': 'none', 'height': 'auto'})

      self <= titlebar
      self <= self.body

      for _i in ['n', 'e', 's', 'w', 'se', 'sw', 'ne', 'nw']:
          if _i == 'se':
             _class="ui-resizable-handle ui-resizable-%s ui-icon ui-icon-gripsmall-diagonal-%s" % (_i, _i)
          else:
             _class="ui-resizable-handle ui-resizable-%s" % _i

          self <= html.DIV(Class=_class, style={'z-index': '90'})

      self.footer = html.DIV(Class='ui-footer')
      self <= self.footer

    def add_ok_cancel(self, ok=None, cancel=None):
        """Add Ok and Cancel buttons
        ok is the callback function if user clicks "Ok". It is a function that
        takes the Dialog instance as single argument. Same for cancel
        """
        self.ok = ok
        self.cancel = cancel

        self.footer.style.height = '1.5em'
        left = html.DIV(style=dict(width="50%"),Class='ui-footer-ok')
        right = html.DIV(style=dict(width="50%"), Class='ui-footer-cancel')
        self.footer <= left+right
        
        ok_button = html.BUTTON('Ok', Class="ui-footer-button")
        if ok is None:
            ok_button.bind('click', lambda ev: self.close())
        else:
            ok_button.bind('click', lambda ev: self.callback(ok))
        left <= ok_button

        cancel_button = html.BUTTON('Annuler', Class="ui-footer-button")
        if cancel is None:
            cancel_button.bind('click', lambda ev: self.close())
        else:
            cancel_button.bind('click', lambda ev: self.callback(cancel))
        right <= cancel_button

    def close(self):
        print('close', self)
        self.parent.remove(self)

    def callback(self, func):
        func(self)

class EntryDialog(Dialog):

    def __init__(self, title, prompt, callback):
        Dialog.__init__(self, title)

        d_prompt = html.DIV(prompt, Class="ui-widget", 
            style=dict(float="left",paddingRight="10px"))
        self.entry = html.INPUT()
        self.body <= html.DIV(d_prompt+self.entry,
            style={'padding':'15px'})
        self.entry.focus()
        self.entry.bind('keydown', self.keydown)
        self.add_ok_cancel(ok=callback)
        self.callback = callback
    
    def keydown(self, ev):
        ev.stopPropagation()

class SelectDialog(Dialog):

    def __init__(self, title, prompt, options, action, _id=None):
        Dialog.__init__(self, _id)
        self.set_title(title)
        self.options = options
        self.action = action
        d_prompt = html.DIV(prompt, Class="ui-widget", 
            style=dict(float="left",paddingRight="10px"))
        self.select = html.SELECT()
        for option in options:
            self.select <= html.OPTION(option)
        body = html.DIV(d_prompt+self.select,
            style={'padding':'15px'})
        b_ok = html.BUTTON("Ok")
        b_ok.bind('click', self.ok)
        b_cancel = html.BUTTON("Cancel")
        b_cancel.bind('click', self.cancel)
        body += html.DIV(b_ok+b_cancel, style={'padding':'15px'})
        self._div_dialog <= body
    
    def ok(self, ev):
        ix = self._div_shell.get(selector='SELECT')[0].selectedIndex
        document.remove(self._div_shell)
        self.action(self.options[ix])

    def cancel(self, ev):
        document.remove(self._div_shell)

class YesNoDialog(Dialog):

    def __init__(self, title, prompt, action_if_yes, action_if_no, _id=None):
        Dialog.__init__(self, _id)
        self.set_title(title)
        
        self.action_if_yes = action_if_yes
        self.action_if_no = action_if_no
        
        d_prompt = html.DIV(prompt, Class="ui-widget", 
            style=dict(float="left",paddingRight="10px"))
        body = html.DIV(d_prompt, style={'padding':'15px'})
        b_ok = html.BUTTON("Yes")
        b_ok.bind('click', self.yes)
        b_cancel = html.BUTTON("No")
        b_cancel.bind('click', self.no)
        body += html.DIV(b_ok+b_cancel, style={'padding':'15px'})
        self._div_dialog <= body
    
    def yes(self, ev):
        document.remove(self._div_shell)
        self.action_if_yes(self)

    def no(self, ev):
        document.remove(self._div_shell)
        if self.action_if_no is not None:
            self.action_if_no(self)



# local storage in browser
import sys
from javascript import JSObject
from browser import window, console

has_local_storage = hasattr(window, 'localStorage')

class __UnProvided():
    pass

class LocalStorage():
    storage_type = "local_storage"

    def __init__(self):
        if not has_local_storage:
            raise EnvironmentError("LocalStorage not available")
        self.store = JSObject(window.localStorage)

    def __delitem__(self, key):
        if (not isinstance(key, str)):
            raise TypeError("key must be string")
        if key not in self:
            raise KeyError(key)
        self.store.removeItem(key)

    def __getitem__(self, key):
        if (not isinstance(key, str)):
            raise TypeError("key must be string")
        res = __BRYTHON__.JSObject(self.store.getItem(key))
        if res is not None:
            return res
        raise KeyError(key)

    def __setitem__(self, key, value):
        if (not isinstance(key, str)):
            raise TypeError("key must be string")
        if (not isinstance(value, str)):
            raise TypeError("value must be string")
        self.store.setItem(key, value)

    # implement "in" functionality
    def __contains__(self, key):
        if (not isinstance(key, str)):
            raise TypeError("key must be string")
        res = __BRYTHON__.JSObject(self.store.getItem(key))
        if res is None:
            return False
        return True

    def __iter__(self):
        keys = self.keys()
        return keys.__iter__()

    def get(self, key, default=None):
        if (not isinstance(key, str)):
            raise TypeError("key must be string")
        return __BRYTHON__.JSObject(self.store.getItem(key)) or default

    def pop(self, key, default=__UnProvided()):
        if (not isinstance(key, str)):
            raise TypeError("key must be string")
        if type(default) is __UnProvided:
            ret = self.get(key)
            del self[key]  # will throw key error if doesn't exist
            return ret
        else:
            if key in self:
                ret = self.get(key)
                del self[key]
                return ret
            else:
                return default

    # while a real dict provides a view, returning a generator would less helpful than simply returning a list
    # and creating a custom iterator is overkill and would likely result in slower performance
    def keys(self):
        return [__BRYTHON__.JSObject(self.store.key(i)) for i in range(self.store.length)]

    def values(self):
        return [__BRYTHON__.JSObject(self.__getitem__(k)) for k in self.keys()]

    def items(self):
        return list(zip(self.keys(), self.values()))

    def clear(self):
        self.store.clear()

    def __len__(self):
        return self.store.length

if has_local_storage:
    storage = LocalStorage()



import sys
import os
from setuptools import setup, find_packages
from regulome_app.webapp.tools.metadata import __author__, __version__, __description__, __package_name__


if sys.version_info < (3, 5):
    raise RuntimeError('This package requires Python 3.5 or later')


try:
    from pypandoc import convert_file

    def read_md(f):
        """Convert Markdown to RST"""
        return convert_file(f, 'rst')
except ImportError:
    convert = None
    print("WARNING: pypandoc module not found, could not convert Markdown to RST")

    def read_md(f):
        return open(f, 'r').read()

README = os.path.join(os.path.dirname(__file__), 'README.md')


setup(
    name=__package_name__,
    version=__version__,
    packages=find_packages(),
    license='MIT License',
    author=__author__,
    author_email='loris@mularoni.com',
    package_data={
        'regulome_web': [
            'regulome_app/webapp/templates/*.html',
            'regulome_app/webapp/static/*/*',
            'regulome_app/tools/*.py',
            'regulome_app/webapp/*.py',
            'regulome_app/regulome.cfg.template'
        ]
    },
    include_package_data=True,
    url="https://bitbucket.org/batterio/regulome_web",
    download_url="https://bitbucket.org/batterio/regulome_web/get/" + __version__ + ".tar.gz",
    description=__description__,
    long_description=read_md(README),
    install_requires=[
        'Flask',
        'Jinja2',
        'Click',
        # 'SQLAlchemy',
        # 'Flask-SQLAlchemy',
        # 'MarkupSafe',
    ],
    entry_points={
        'console_scripts': [
            'regulome_web = regulome_app.run:main'
        ]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
    ],
)





import sys
from browser import console

def print_exc(file=sys.stderr):
    exc = __BRYTHON__.current_exception
    if isinstance(exc, SyntaxError):
        file.write('\n module %s line %s' %(exc.args[1], exc.args[2]))
        offset = exc.args[3]
        file.write('\n  '+exc.args[4])
        file.write('\n  '+offset*' '+'^')
    else:
        file.write(exc.info)
    file.write('\n'+exc.__name__)
    if exc.args:
        file.write(': %s' %exc.args[0])
    file.write('\n')

def format_exc(limit=None,chain=True):
    exc = __BRYTHON__.current_exception
    res = exc.info+'\n'+exc.__name__
    if exc.args:
        res += ': '+exc.args[0]
    return res+'\n'

def format_exception(_type, value, tb, limit=None, chain=True):
    return ['%s\n' %_type,'%s\n' %value]    

def extract_tb(tb, limit=None):
    return tb


"""CSV parsing and writing.

[Copied from PyPy
https://bitbucket-assetroot.s3.amazonaws.com/pypy/pypy/1400171824.19/641/_csv.py?Signature=cc%2Bc8m06cBMbsxt2e15XXXUDACk%3D&Expires=1404136251&AWSAccessKeyId=0EMWEFSGA12Z1HF1TZ82
and adapted to Python 3 syntax for Brython]


This module provides classes that assist in the reading and writing
of Comma Separated Value (CSV) files, and implements the interface
described by PEP 305.  Although many CSV files are simple to parse,
the format is not formally defined by a stable specification and
is subtle enough that parsing lines of a CSV file with something
like line.split(\",\") is bound to fail.  The module supports three
basic APIs: reading, writing, and registration of dialects.


DIALECT REGISTRATION:

Readers and writers support a dialect argument, which is a convenient
handle on a group of settings.  When the dialect argument is a string,
it identifies one of the dialects previously registered with the module.
If it is a class or instance, the attributes of the argument are used as
the settings for the reader or writer:

    class excel:
        delimiter = ','
        quotechar = '\"'
        escapechar = None
        doublequote = True
        skipinitialspace = False
        lineterminator = '\\r\\n'
        quoting = QUOTE_MINIMAL

SETTINGS:

    * quotechar - specifies a one-character string to use as the 
        quoting character.  It defaults to '\"'.
    * delimiter - specifies a one-character string to use as the 
        field separator.  It defaults to ','.
    * skipinitialspace - specifies how to interpret whitespace which
        immediately follows a delimiter.  It defaults to False, which
        means that whitespace immediately following a delimiter is part
        of the following field.
    * lineterminator -  specifies the character sequence which should 
        terminate rows.
    * quoting - controls when quotes should be generated by the writer.
        It can take on any of the following module constants:

        csv.QUOTE_MINIMAL means only when required, for example, when a
            field contains either the quotechar or the delimiter
        csv.QUOTE_ALL means that quotes are always placed around fields.
        csv.QUOTE_NONNUMERIC means that quotes are always placed around
            fields which do not parse as integers or floating point
            numbers.
        csv.QUOTE_NONE means that quotes are never placed around fields.
    * escapechar - specifies a one-character string used to escape 
        the delimiter when quoting is set to QUOTE_NONE.
    * doublequote - controls the handling of quotes inside fields.  When
        True, two consecutive quotes are interpreted as one during read,
        and when writing, each quote character embedded in the data is
        written as two quotes.
"""

__version__ = "1.0"

QUOTE_MINIMAL, QUOTE_ALL, QUOTE_NONNUMERIC, QUOTE_NONE = range(4)
_dialects = {}
_field_limit = 128 * 1024 # max parsed field size

class Error(Exception):
    pass

class Dialect(object):
    """CSV dialect

    The Dialect type records CSV parsing and generation options."""

    __slots__ = ["_delimiter", "_doublequote", "_escapechar",
                 "_lineterminator", "_quotechar", "_quoting",
                 "_skipinitialspace", "_strict"]

    def __new__(cls, dialect, **kwargs):

        for name in kwargs:
            if '_' + name not in Dialect.__slots__:
                raise TypeError("unexpected keyword argument '%s'" %
                                (name,))

        if dialect is not None:
            if isinstance(dialect, str):
                dialect = get_dialect(dialect)

            # Can we reuse this instance?
            if (isinstance(dialect, Dialect)
                and all(value is None for value in kwargs.values())):
                return dialect

        self = object.__new__(cls)


        def set_char(x):
            if x is None:
                return None
            if isinstance(x, str) and len(x) <= 1:
                return x
            raise TypeError("%r must be a 1-character string" % (name,))
        def set_str(x):
            if isinstance(x, str):
                return x
            raise TypeError("%r must be a string" % (name,))
        def set_quoting(x):
            if x in range(4):
                return x
            raise TypeError("bad 'quoting' value")

        attributes = {"delimiter": (',', set_char),
                      "doublequote": (True, bool),
                      "escapechar": (None, set_char),
                      "lineterminator": ("\r\n", set_str),
                      "quotechar": ('"', set_char),
                      "quoting": (QUOTE_MINIMAL, set_quoting),
                      "skipinitialspace": (False, bool),
                      "strict": (False, bool),
                      }

        # Copy attributes
        notset = object()
        for name in Dialect.__slots__:
            name = name[1:]
            value = notset
            if name in kwargs:
                value = kwargs[name]
            elif dialect is not None:
                value = getattr(dialect, name, notset)

            # mapping by name: (default, converter)
            if value is notset:
                value = attributes[name][0]
                if name == 'quoting' and not self.quotechar:
                    value = QUOTE_NONE
            else:
                converter = attributes[name][1]
                if converter:
                    value = converter(value)

            setattr(self, '_' + name, value)

        if not self.delimiter:
            raise TypeError("delimiter must be set")

        if self.quoting != QUOTE_NONE and not self.quotechar:
            raise TypeError("quotechar must be set if quoting enabled")

        if not self.lineterminator:
            raise TypeError("lineterminator must be set")

        return self

    delimiter        = property(lambda self: self._delimiter)
    doublequote      = property(lambda self: self._doublequote)
    escapechar       = property(lambda self: self._escapechar)
    lineterminator   = property(lambda self: self._lineterminator)
    quotechar        = property(lambda self: self._quotechar)
    quoting          = property(lambda self: self._quoting)
    skipinitialspace = property(lambda self: self._skipinitialspace)
    strict           = property(lambda self: self._strict)


def _call_dialect(dialect_inst, kwargs):
    return Dialect(dialect_inst, **kwargs)

def register_dialect(name, dialect=None, **kwargs):
    """Create a mapping from a string name to a dialect class.
    dialect = csv.register_dialect(name, dialect)"""
    if not isinstance(name, str):
        raise TypeError("dialect name must be a string or unicode")

    dialect = _call_dialect(dialect, kwargs)
    _dialects[name] = dialect

def unregister_dialect(name):
    """Delete the name/dialect mapping associated with a string name.\n
    csv.unregister_dialect(name)"""
    try:
        del _dialects[name]
    except KeyError:
        raise Error("unknown dialect")

def get_dialect(name):
    """Return the dialect instance associated with name.
    dialect = csv.get_dialect(name)"""
    try:
        return _dialects[name]
    except KeyError:
        raise Error("unknown dialect")

def list_dialects():
    """Return a list of all know dialect names
    names = csv.list_dialects()"""
    return list(_dialects)

class Reader(object):
    """CSV reader

    Reader objects are responsible for reading and parsing tabular data
    in CSV format."""


    (START_RECORD, START_FIELD, ESCAPED_CHAR, IN_FIELD,
     IN_QUOTED_FIELD, ESCAPE_IN_QUOTED_FIELD, QUOTE_IN_QUOTED_FIELD,
     EAT_CRNL) = range(8)

    def __init__(self, iterator, dialect=None, **kwargs):
        self.dialect = _call_dialect(dialect, kwargs)

        # null characters are not allowed to be in the string so we can use
        # it as a fall back
        self._delimiter = self.dialect.delimiter if self.dialect.delimiter else '\0'
        self._quotechar = self.dialect.quotechar if self.dialect.quotechar else '\0'
        self._escapechar = self.dialect.escapechar if self.dialect.escapechar else '\0'
        self._doublequote = self.dialect.doublequote
        self._quoting = self.dialect.quoting
        self._skipinitialspace = self.dialect.skipinitialspace
        self._strict = self.dialect.strict

        self.input_iter = iter(iterator)
        self.line_num = 0

        self._parse_reset()

    def _parse_reset(self):
        self.field = ''
        self.fields = []
        self.state = self.START_RECORD
        self.numeric_field = False

    def __iter__(self):
        return self

    def __next__(self):
        self._parse_reset()
        while True:
            try:
                line = next(self.input_iter)
            except StopIteration:
                # End of input OR exception
                if len(self.field) > 0:
                    raise Error("newline inside string")
                raise

            self.line_num += 1

            if '\0' in line:
                raise Error("line contains NULL byte")
            self._parse_process_char(line)
            self._parse_eol()

            if self.state == self.START_RECORD:
                break

        fields = self.fields
        self.fields = []
        return fields

    def _parse_process_char(self, line):
        pos = 0
        while pos < len(line):
            if self.state == self.IN_FIELD:
                # in unquoted field and have already found one character when starting the field
                pos2 = pos
                while pos2 < len(line):
                    if line[pos2] == '\n' or line[pos2] == '\r':
                        # end of line - return [fields]
                        if pos2 > pos:
                            self._parse_add_str(line[pos:pos2])
                            pos = pos2
                        self._parse_save_field()
                        self.state = self.EAT_CRNL
                        break
                    elif line[pos2] == self._escapechar[0]:
                        # possible escaped character
                        if pos2 > pos:
                            self._parse_add_str(line[pos:pos2])
                        pos = pos2
                        self.state = self.ESCAPED_CHAR
                        break
                    elif line[pos2] == self._delimiter[0]:
                        # save field - wait for new field
                        if pos2 > pos:
                            self._parse_add_str(line[pos:pos2])
                            pos = pos2
                        self._parse_save_field()
                        self.state = self.START_FIELD
                        break
                    # normal character - save in field
                    pos2 += 1
                else:
                    if pos2 > pos:
                        self._parse_add_str(line[pos:pos2])
                        pos = pos2
                        continue

            elif self.state == self.START_RECORD:
                if line[pos] == '\n' or line[pos] == '\r':
                    self.state = self.EAT_CRNL
                else:
                    self.state = self.START_FIELD
                    # restart process
                    continue

            elif self.state == self.START_FIELD:
                if line[pos] == '\n' or line[pos] == '\r':
                    # save empty field - return [fields]
                    self._parse_save_field()
                    self.state = self.EAT_CRNL
                elif (line[pos] == self._quotechar[0]
                      and self._quoting != QUOTE_NONE):
                    # start quoted field
                    self.state = self.IN_QUOTED_FIELD
                elif line[pos] == self._escapechar[0]:
                    # possible escaped character
                    self.state = self.ESCAPED_CHAR
                elif self._skipinitialspace and line[pos] == ' ':
                    # ignore space at start of field
                    pass
                elif line[pos] == self._delimiter[0]:
                    # save empty field
                    self._parse_save_field()
                else:
                    # begin new unquoted field
                    if self._quoting == QUOTE_NONNUMERIC:
                        self.numeric_field = True
                    self.state = self.IN_FIELD
                    continue

            elif self.state == self.ESCAPED_CHAR:
                self._parse_add_char(line[pos])
                self.state = self.IN_FIELD

            elif self.state == self.IN_QUOTED_FIELD:
                if line[pos] == self._escapechar:
                    # possible escape character
                    self.state = self.ESCAPE_IN_QUOTED_FIELD
                elif (line[pos] == self._quotechar
                      and self._quoting != QUOTE_NONE):
                    if self._doublequote:
                        # doublequote; " represented by ""
                        self.state = self.QUOTE_IN_QUOTED_FIELD
                    else:
                        #end of quote part of field
                        self.state = self.IN_FIELD
                else:
                    # normal character - save in field
                    self._parse_add_char(line[pos])

            elif self.state == self.ESCAPE_IN_QUOTED_FIELD:
                self._parse_add_char(line[pos])
                self.state = self.IN_QUOTED_FIELD

            elif self.state == self.QUOTE_IN_QUOTED_FIELD:
                # doublequote - seen a quote in a quoted field
                if (line[pos] == self._quotechar
                    and self._quoting != QUOTE_NONE):
                    # save "" as "
                    self._parse_add_char(line[pos])
                    self.state = self.IN_QUOTED_FIELD
                elif line[pos] == self._delimiter[0]:
                    # save field - wait for new field
                    self._parse_save_field()
                    self.state = self.START_FIELD
                elif line[pos] == '\r' or line[pos] == '\n':
                    # end of line - return [fields]
                    self._parse_save_field()
                    self.state = self.EAT_CRNL
                elif not self._strict:
                    self._parse_add_char(line[pos])
                    self.state = self.IN_FIELD
                else:
                    raise Error("'%c' expected after '%c'" %
                                (self._delimiter, self._quotechar))

            elif self.state == self.EAT_CRNL:
                if line[pos] == '\r' or line[pos] == '\n':
                    pass
                else:
                    raise Error("new-line character seen in unquoted field - "
                                "do you need to open the file "
                                "in universal-newline mode?")

            else:
                raise RuntimeError("unknown state: %r" % (self.state,))

            pos += 1

    def _parse_eol(self):
        if self.state == self.EAT_CRNL:
            self.state = self.START_RECORD
        elif self.state == self.START_RECORD:
            # empty line - return []
            pass
        elif self.state == self.IN_FIELD:
            # in unquoted field
            # end of line - return [fields]
            self._parse_save_field()
            self.state = self.START_RECORD
        elif self.state == self.START_FIELD:
            # save empty field - return [fields]
            self._parse_save_field()
            self.state = self.START_RECORD
        elif self.state == self.ESCAPED_CHAR:
            self._parse_add_char('\n')
            self.state = self.IN_FIELD
        elif self.state == self.IN_QUOTED_FIELD:
            pass
        elif self.state == self.ESCAPE_IN_QUOTED_FIELD:
            self._parse_add_char('\n')
            self.state = self.IN_QUOTED_FIELD
        elif self.state == self.QUOTE_IN_QUOTED_FIELD:
            # end of line - return [fields]
            self._parse_save_field()
            self.state = self.START_RECORD
        else:
            raise RuntimeError("unknown state: %r" % (self.state,))

    def _parse_save_field(self):
        field, self.field = self.field, ''
        if self.numeric_field:
            self.numeric_field = False
            field = float(field)
        self.fields.append(field)

    def _parse_add_char(self, c):
        if len(self.field) + 1 > _field_limit:
            raise Error("field larget than field limit (%d)" % (_field_limit))
        self.field += c

    def _parse_add_str(self, s):
        if len(self.field) + len(s) > _field_limit:
            raise Error("field larget than field limit (%d)" % (_field_limit))
        self.field += s


class Writer(object):
    """CSV writer

    Writer objects are responsible for generating tabular data
    in CSV format from sequence input."""

    def __init__(self, file, dialect=None, **kwargs):
        if not (hasattr(file, 'write') and callable(file.write)):
            raise TypeError("argument 1 must have a 'write' method")
        self.writeline = file.write
        self.dialect = _call_dialect(dialect, kwargs)

    def _join_reset(self):
        self.rec = []
        self.num_fields = 0

    def _join_append(self, field, quoted, quote_empty):
        dialect = self.dialect
        # If this is not the first field we need a field separator
        if self.num_fields > 0:
            self.rec.append(dialect.delimiter)

        if dialect.quoting == QUOTE_NONE:
            need_escape = tuple(dialect.lineterminator) + (
                dialect.escapechar,  # escapechar always first
                dialect.delimiter, dialect.quotechar)

        else:
            for c in tuple(dialect.lineterminator) + (
                dialect.delimiter, dialect.escapechar):
                if c and c in field:
                    quoted = True

            need_escape = ()
            if dialect.quotechar in field:
                if dialect.doublequote:
                    field = field.replace(dialect.quotechar,
                                          dialect.quotechar * 2)
                    quoted = True
                else:
                    need_escape = (dialect.quotechar,)


        for c in need_escape:
            if c and c in field:
                if not dialect.escapechar:
                    raise Error("need to escape, but no escapechar set")
                field = field.replace(c, dialect.escapechar + c)

        # If field is empty check if it needs to be quoted
        if field == '' and quote_empty:
            if dialect.quoting == QUOTE_NONE:
                raise Error("single empty field record must be quoted")
            quoted = 1

        if quoted:
            field = dialect.quotechar + field + dialect.quotechar

        self.rec.append(field)
        self.num_fields += 1



    def writerow(self, row):
        dialect = self.dialect
        try:
            rowlen = len(row)
        except TypeError:
            raise Error("sequence expected")

        # join all fields in internal buffer
        self._join_reset()

        for field in row:
            quoted = False
            if dialect.quoting == QUOTE_NONNUMERIC:
                try:
                    float(field)
                except:
                    quoted = True
                # This changed since 2.5:
                # quoted = not isinstance(field, (int, long, float))
            elif dialect.quoting == QUOTE_ALL:
                quoted = True

            if field is None:
                self._join_append("", quoted, rowlen == 1)
            else:
                self._join_append(str(field), quoted, rowlen == 1)

        # add line terminator
        self.rec.append(dialect.lineterminator)

        self.writeline(''.join(self.rec))

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)

def reader(*args, **kwargs):
    """
    csv_reader = reader(iterable [, dialect='excel']
                       [optional keyword args])
    for row in csv_reader:
        process(row)

    The "iterable" argument can be any object that returns a line
    of input for each iteration, such as a file object or a list.  The
    optional \"dialect\" parameter is discussed below.  The function
    also accepts optional keyword arguments which override settings
    provided by the dialect.

    The returned object is an iterator.  Each iteration returns a row
    of the CSV file (which can span multiple input lines)"""

    return Reader(*args, **kwargs)

def writer(*args, **kwargs):
    """
    csv_writer = csv.writer(fileobj [, dialect='excel']
                            [optional keyword args])
    for row in sequence:
        csv_writer.writerow(row)

    [or]

    csv_writer = csv.writer(fileobj [, dialect='excel']
                            [optional keyword args])
    csv_writer.writerows(rows)

    The \"fileobj\" argument can be any object that supports the file API."""
    return Writer(*args, **kwargs)


undefined = object()
def field_size_limit(limit=undefined):
    """Sets an upper limit on parsed fields.
    csv.field_size_limit([limit])

    Returns old limit. If limit is not given, no new limit is set and
    the old limit is returned"""

    global _field_limit
    old_limit = _field_limit

    if limit is not undefined:
        if not isinstance(limit, (int, long)):
            raise TypeError("int expected, got %s" %
                            (limit.__class__.__name__,))
        _field_limit = limit

    return old_limit



class EventListener:
  def __init__(self, events=[]):
      self._events=events

  def append(self, event):
      self._events.append(event)

  def fire(self, e):
      for _event in self._events:
          _event(e)

class IndexedDB:
  def __init__(self):
      if not __BRYTHON__.has_indexedDB:
         raise NotImplementedError("Your browser doesn't support indexedDB")
         return

      self._indexedDB=__BRYTHON__.indexedDB()
      self._db=None
      self._version=None

  def _onsuccess(self, event):
      self._db=event.target.result

  def open(self, name, onsuccess, version=1.0, onerror=None, 
           onupgradeneeded=None):
      self._version=version
      _result=self._indexedDB.open(name, version)
      _success=EventListener([self._onsuccess, onsuccess])
      _result.onsuccess=_success.fire
      _result.onupgradeneeded=onupgradeneeded

      #if onerror is None:
      def onerror(e):
          print("onerror: %s:%s" %  (e.type, e.target.result))

      def onblocked(e):
          print("blocked: %s:%s" %  (e.type, e.result))

      _result.onerror=onerror
      _result.onblocked=onblocked

  def transaction(self, entities, mode='read'):
      return Transaction(self._db.transaction(entities, mode))

class Transaction:

  def __init__(self, transaction):
      self._transaction=transaction

  def objectStore(self, name):
      return ObjectStore(self._transaction.objectStore(name))

class ObjectStore:

  def __init__(self, objectStore):
      self._objectStore=objectStore
      self._data=[]

  def clear(self, onsuccess=None, onerror=None):
      _result=self._objectStore.clear()

      if onsuccess is not None:
         _result.onsuccess=onsuccess

      if onerror is not None:
         _result.onerror=onerror

  def _helper(self, func, object, onsuccess=None, onerror=None):
      _result=func(object)

      if onsuccess is not None:
         _result.onsuccess=onsuccess

      if onerror is not None:
         _result.onerror=onerror

  def put(self, obj, key=None, onsuccess=None, onerror=None):
      _r = self._objectStore.put(obj, key)
      _r.onsuccess = onsuccess
      _r.onerror = onerror

  def add(self, obj, key, onsuccess=None, onerror=None):
      _r = self._objectStore.add(obj, key)
      _r.onsuccess = onsuccess
      _r.onerror = onerror
      #self._helper(self._objectStore.add, object, onsuccess, onerror)

  def delete(self, index, onsuccess=None, onerror=None): 
      self._helper(self._objectStore.delete, index, onsuccess, onerror)
     
  def query(self, *args):
      self._data=[]
      def onsuccess(event):
          cursor=event.target.result
          if cursor is not None:
             self._data.append(cursor.value)
             getattr(cursor,"continue")() # cursor.continue() is illegal

      self._objectStore.openCursor(args).onsuccess=onsuccess

  def fetchall(self):
      yield self._data

  def get(self, key, onsuccess=None, onerror=None):
      self._helper(self._objectStore.get, key, onsuccess, onerror)



__author__ = 'Loris Mularoni'


# Import modules
from flask import Flask
# from flask_sqlalchemy import SQLAlchemy


# Create the application
app = Flask(__name__)
# db = SQLAlchemy(app)


import regulome_app.webapp.views



class ProxyType:

    def __init__(self,obj):
        self.obj = obj

CallableProxyType = ProxyType
ProxyTypes = [ProxyType,CallableProxyType]

class ReferenceType:

    def __init__(self,obj,callback):
        self.obj = obj
        self.callback = callback

class ref:

    def __init__(self,obj,callback=None):
        self.obj = ReferenceType(obj,callback)
        self.callback=callback

    def __call__(self):
        return self.obj.obj

def getweakrefcount(obj):
    return 1

def getweakrefs(obj):
    return obj


def proxy(obj,callback=None):
    return ProxyType(obj)




#
# This module is a pure Python version of pypy.module.struct.
# It is only imported if the vastly faster pypy.module.struct is not
# compiled in.  For now we keep this version for reference and
# because pypy.module.struct is not ootype-backend-friendly yet.
#

# this module 'borrowed' from 
# https://bitbucket.org/pypy/pypy/src/18626459a9b2/lib_pypy/_struct.py?at=py3k-listview_str
# with many bug fixes

"""Functions to convert between Python values and C structs.
Python strings are used to hold the data representing the C struct
and also as format strings to describe the layout of data in the C struct.

The optional first format char indicates byte order, size and alignment:
 @: native order, size & alignment (default)
 =: native order, std. size & alignment
 <: little-endian, std. size & alignment
 >: big-endian, std. size & alignment
 !: same as >

The remaining chars indicate types of args and must match exactly;
these can be preceded by a decimal repeat count:
   x: pad byte (no data);
   c:char;
   b:signed byte;
   B:unsigned byte;
   h:short;
   H:unsigned short;
   i:int;
   I:unsigned int;
   l:long;
   L:unsigned long;
   f:float;
   d:double.
Special cases (preceding decimal count indicates length):
   s:string (array of char); p: pascal string (with count byte).
Special case (only available in native format):
   P:an integer type that is wide enough to hold a pointer.
Special case (not in native mode unless 'long long' in platform C):
   q:long long;
   Q:unsigned long long
Whitespace between formats is ignored.

The variable struct.error is an exception raised on errors."""

import math, sys

# TODO: XXX Find a way to get information on native sizes and alignments
class StructError(Exception):
    pass
error = StructError
def unpack_int(data,index,size,le):
    bytes = [b for b in data[index:index+size]]
    if le == 'little':
        bytes.reverse()
    number = 0
    for b in bytes:
        number = number << 8 | b
    return int(number)

def unpack_signed_int(data,index,size,le):
    number = unpack_int(data,index,size,le)
    max = 2**(size*8)
    if number > 2**(size*8 - 1) - 1:
        number = int(-1*(max - number))
    return number

INFINITY = 1e200 * 1e200
NAN = INFINITY / INFINITY

def unpack_char(data,index,size,le):
    return data[index:index+size]

def pack_int(number,size,le):
    x=number
    res=[]
    for i in range(size):
        res.append(x&0xff)
        x >>= 8
    if le == 'big':
        res.reverse()
    return bytes(res)

def pack_signed_int(number,size,le):
    if not isinstance(number, int):
        raise StructError("argument for i,I,l,L,q,Q,h,H must be integer")
    if  number > 2**(8*size-1)-1 or number < -1*2**(8*size-1):
        raise OverflowError("Number:%i too large to convert" % number)
    return pack_int(number,size,le)

def pack_unsigned_int(number,size,le):
    if not isinstance(number, int):
        raise StructError("argument for i,I,l,L,q,Q,h,H must be integer")
    if number < 0:
        raise TypeError("can't convert negative long to unsigned")
    if number > 2**(8*size)-1:
        raise OverflowError("Number:%i too large to convert" % number)
    return pack_int(number,size,le)

def pack_char(char,size,le):
    return bytes(char)

def isinf(x):
    return x != 0.0 and x / 2 == x
def isnan(v):
    return v != v*1.0 or (v == 1.0 and v == 2.0)

def pack_float(x, size, le):
    unsigned = float_pack(x, size)
    result = []
    for i in range(size):
        result.append((unsigned >> (i * 8)) & 0xFF)
    if le == "big":
        result.reverse()
    return bytes(result)

def unpack_float(data, index, size, le):
    binary = [data[i] for i in range(index, index + size)]
    if le == "big":
        binary.reverse()
    unsigned = 0
    for i in range(size):
        unsigned |= binary[i] << (i * 8)
    return float_unpack(unsigned, size, le)

def round_to_nearest(x):
    """Python 3 style round:  round a float x to the nearest int, but
    unlike the builtin Python 2.x round function:

      - return an int, not a float
      - do round-half-to-even, not round-half-away-from-zero.

    We assume that x is finite and nonnegative; except wrong results
    if you use this for negative x.

    """
    int_part = int(x)
    frac_part = x - int_part
    if frac_part > 0.5 or frac_part == 0.5 and int_part & 1 == 1:
        int_part += 1
    return int_part

def float_unpack(Q, size, le):
    """Convert a 32-bit or 64-bit integer created
    by float_pack into a Python float."""

    if size == 8:
        MIN_EXP = -1021  # = sys.float_info.min_exp
        MAX_EXP = 1024   # = sys.float_info.max_exp
        MANT_DIG = 53    # = sys.float_info.mant_dig
        BITS = 64
    elif size == 4:
        MIN_EXP = -125   # C's FLT_MIN_EXP
        MAX_EXP = 128    # FLT_MAX_EXP
        MANT_DIG = 24    # FLT_MANT_DIG
        BITS = 32
    else:
        raise ValueError("invalid size value")

    if Q >> BITS:
         raise ValueError("input out of range")

    # extract pieces
    sign = Q >> BITS - 1
    exp = (Q & ((1 << BITS - 1) - (1 << MANT_DIG - 1))) >> MANT_DIG - 1
    mant = Q & ((1 << MANT_DIG - 1) - 1)

    if exp == MAX_EXP - MIN_EXP + 2:
        # nan or infinity
        result = float('nan') if mant else float('inf')
    elif exp == 0:
        # subnormal or zero
        result = math.ldexp(float(mant), MIN_EXP - MANT_DIG)
    else:
        # normal
        mant += 1 << MANT_DIG - 1
        result = math.ldexp(float(mant), exp + MIN_EXP - MANT_DIG - 1)
    return -result if sign else result


def float_pack(x, size):
    """Convert a Python float x into a 64-bit unsigned integer
    with the same byte representation."""

    if size == 8:
        MIN_EXP = -1021  # = sys.float_info.min_exp
        MAX_EXP = 1024   # = sys.float_info.max_exp
        MANT_DIG = 53    # = sys.float_info.mant_dig
        BITS = 64
    elif size == 4:
        MIN_EXP = -125   # C's FLT_MIN_EXP
        MAX_EXP = 128    # FLT_MAX_EXP
        MANT_DIG = 24    # FLT_MANT_DIG
        BITS = 32
    else:
        raise ValueError("invalid size value")

    sign = math.copysign(1.0, x) < 0.0
    if math.isinf(x):
        mant = 0
        exp = MAX_EXP - MIN_EXP + 2
    elif math.isnan(x):
        mant = 1 << (MANT_DIG-2) # other values possible
        exp = MAX_EXP - MIN_EXP + 2
    elif x == 0.0:
        mant = 0
        exp = 0
    else:
        m, e = math.frexp(abs(x))  # abs(x) == m * 2**e
        exp = e - (MIN_EXP - 1)
        if exp > 0:
            # Normal case.
            mant = round_to_nearest(m * (1 << MANT_DIG))
            mant -= 1 << MANT_DIG - 1
        else:
            # Subnormal case.
            if exp + MANT_DIG - 1 >= 0:
                mant = round_to_nearest(m * (1 << exp + MANT_DIG - 1))
            else:
                mant = 0
            exp = 0

        # Special case: rounding produced a MANT_DIG-bit mantissa.
        assert 0 <= mant <= 1 << MANT_DIG - 1
        if mant == 1 << MANT_DIG - 1:
            mant = 0
            exp += 1

        # Raise on overflow (in some circumstances, may want to return
        # infinity instead).
        if exp >= MAX_EXP - MIN_EXP + 2:
             raise OverflowError("float too large to pack in this format")

    # check constraints
    assert 0 <= mant < 1 << MANT_DIG - 1
    assert 0 <= exp <= MAX_EXP - MIN_EXP + 2
    assert 0 <= sign <= 1
    return ((sign << BITS - 1) | (exp << MANT_DIG - 1)) | mant


big_endian_format = {
    'x':{ 'size' : 1, 'alignment' : 0, 'pack' : None, 'unpack' : None},
    'b':{ 'size' : 1, 'alignment' : 0, 'pack' : pack_signed_int, 'unpack' : unpack_signed_int},
    'B':{ 'size' : 1, 'alignment' : 0, 'pack' : pack_unsigned_int, 'unpack' : unpack_int},
    'c':{ 'size' : 1, 'alignment' : 0, 'pack' : pack_char, 'unpack' : unpack_char},
    's':{ 'size' : 1, 'alignment' : 0, 'pack' : None, 'unpack' : None},
    'p':{ 'size' : 1, 'alignment' : 0, 'pack' : None, 'unpack' : None},
    'h':{ 'size' : 2, 'alignment' : 0, 'pack' : pack_signed_int, 'unpack' : unpack_signed_int},
    'H':{ 'size' : 2, 'alignment' : 0, 'pack' : pack_unsigned_int, 'unpack' : unpack_int},
    'i':{ 'size' : 4, 'alignment' : 0, 'pack' : pack_signed_int, 'unpack' : unpack_signed_int},
    'I':{ 'size' : 4, 'alignment' : 0, 'pack' : pack_unsigned_int, 'unpack' : unpack_int},
    'l':{ 'size' : 4, 'alignment' : 0, 'pack' : pack_signed_int, 'unpack' : unpack_signed_int},
    'L':{ 'size' : 4, 'alignment' : 0, 'pack' : pack_unsigned_int, 'unpack' : unpack_int},
    'q':{ 'size' : 8, 'alignment' : 0, 'pack' : pack_signed_int, 'unpack' : unpack_signed_int},
    'Q':{ 'size' : 8, 'alignment' : 0, 'pack' : pack_unsigned_int, 'unpack' : unpack_int},
    'f':{ 'size' : 4, 'alignment' : 0, 'pack' : pack_float, 'unpack' : unpack_float},
    'd':{ 'size' : 8, 'alignment' : 0, 'pack' : pack_float, 'unpack' : unpack_float},
    }
default = big_endian_format
formatmode={ '<' : (default, 'little'),
             '>' : (default, 'big'),
             '!' : (default, 'big'),
             '=' : (default, sys.byteorder),
             '@' : (default, sys.byteorder)
            }

def getmode(fmt):
    try:
        formatdef,endianness = formatmode[fmt[0]]
        alignment = fmt[0] not in formatmode or fmt[0]=='@'
        index = 1
    except (IndexError, KeyError):
        formatdef,endianness = formatmode['@']
        alignment = True
        index = 0
    return formatdef,endianness,index,alignment

def getNum(fmt,i):
    num=None
    cur = fmt[i]
    while ('0'<= cur ) and ( cur <= '9'):
        if num == None:
            num = int(cur)
        else:
            num = 10*num + int(cur)
        i += 1
        cur = fmt[i]
    return num,i

def calcsize(fmt):
    """calcsize(fmt) -> int
    Return size of C struct described by format string fmt.
    See struct.__doc__ for more on format strings."""

    formatdef,endianness,i,alignment = getmode(fmt)
    num = 0
    result = 0
    while i<len(fmt):
        num,i = getNum(fmt,i)
        cur = fmt[i]
        try:
            format = formatdef[cur]
        except KeyError:
            raise StructError("%s is not a valid format" % cur)
        if num != None :
            result += num*format['size']
        else:
            # if formatdef is native, alignment is native, so we count a
            # number of padding bytes until result is a multiple of size
            if alignment:
                result += format['size'] - result % format['size']
            result += format['size']
        num = 0
        i += 1
    return result

def pack(fmt,*args):
    """pack(fmt, v1, v2, ...) -> string
       Return string containing values v1, v2, ... packed according to fmt.
       See struct.__doc__ for more on format strings."""
    formatdef,endianness,i,alignment = getmode(fmt)
    args = list(args)
    n_args = len(args)
    result = []
    while i<len(fmt):
        num,i = getNum(fmt,i)
        cur = fmt[i]
        try:
            format = formatdef[cur]
        except KeyError:
            raise StructError("%s is not a valid format" % cur)
        if num == None :
            num_s = 0
            num = 1
        else:
            num_s = num

        if cur == 'x':
            result += [b'\0'*num]
        elif cur == 's':
            if isinstance(args[0], bytes):
                padding = num - len(args[0])
                result += [args[0][:num] + b'\0'*padding]
                args.pop(0)
            else:
                raise StructError("arg for string format not a string")
        elif cur == 'p':
            if isinstance(args[0], bytes):
                padding = num - len(args[0]) - 1

                if padding > 0:
                    result += [bytes([len(args[0])]) + args[0][:num-1] + b'\0'*padding]
                else:
                    if num<255:
                        result += [bytes([num-1]) + args[0][:num-1]]
                    else:
                        result += [bytes([255]) + args[0][:num-1]]
                args.pop(0)
            else:
                raise StructError("arg for string format not a string")

        else:
            if len(args) < num:
                raise StructError("insufficient arguments to pack")
            for var in args[:num]:
                # pad with 0 until position is a multiple of size
                if len(result) and alignment:
                    padding = format['size'] - len(result) % format['size']
                    result += [bytes([0])]*padding
                result += [format['pack'](var,format['size'],endianness)]
            args=args[num:]
        num = None
        i += 1
    if len(args) != 0:
        raise StructError("too many arguments for pack format")
    return b''.join(result)

def unpack(fmt,data):
    """unpack(fmt, string) -> (v1, v2, ...)
       Unpack the string, containing packed C structure data, according
       to fmt.  Requires len(string)==calcsize(fmt).
       See struct.__doc__ for more on format strings."""
    formatdef,endianness,i,alignment = getmode(fmt)
    j = 0
    num = 0
    result = []
    length= calcsize(fmt)
    if length != len (data):
        raise StructError("unpack str size does not match format")
    while i<len(fmt):
        num,i=getNum(fmt,i)
        cur = fmt[i]
        i += 1
        try:
            format = formatdef[cur]
        except KeyError:
            raise StructError("%s is not a valid format" % cur)

        if not num :
            num = 1

        if cur == 'x':
            j += num
        elif cur == 's':
            result.append(data[j:j+num])
            j += num
        elif cur == 'p':
            n=data[j]
            if n >= num:
                n = num-1
            result.append(data[j+1:j+n+1])
            j += num
        else:
            # skip padding bytes until we get at a multiple of size
            if j>0 and alignment:
                padding = format['size'] - j % format['size']
                j += padding
            for n in range(num):
                result += [format['unpack'](data,j,format['size'],endianness)]
                j += format['size']

    return tuple(result)

def pack_into(fmt, buf, offset, *args):
    data = pack(fmt, *args)
    buf[offset:offset+len(data)] = data

def unpack_from(fmt, buf, offset=0):
    size = calcsize(fmt)
    data = buf[offset:offset+size]
    if len(data) != size:
        raise error("unpack_from requires a buffer of at least %d bytes"
                    % (size,))
    return unpack(fmt, data)

def _clearcache():
    "Clear the internal cache."
    # No cache in this implementation

if __name__=='__main__':
    t = pack('Bf',1,2)
    print(t, len(t))
    print(unpack('Bf', t))
    print(calcsize('Bf'))
    
    


from browser import html, document
from .dialog import *
from .progressbar import *
from .slider import *
from .menu import *

def add_stylesheet():
    # should use urllib.parse, but it imports so many modules...
    css_url = '/'.join(__file__.split('/')[:-1]+
        ['css/smoothness/jquery-ui-1.10.3.custom.min.css'])
    _link=html.LINK(Href=css_url)
    _link.rel='stylesheet'

    document <= _link
